@@ -44,13 +44,15 @@
 //! # }
 //! ```
 
+use constant_time_eq::constant_time_eq;
+
 #[cfg(feature = "serde_support")]
 use serde::{Deserialize, Serialize};
 
 use core::fmt;
 
 #[cfg(feature = "qr")]
-use {base64, image::Luma, qrcode::QrCode};
+use {base64, image::Luma, qrcodegen};
 
 use hmac::Mac;
 
@@ -59,7 +61,7 @@ type HmacSha256 = hmac::Hmac<sha2::Sha256>;
 type HmacSha512 = hmac::Hmac<sha2::Sha512>;
 
 /// Algorithm enum holds the three standards algorithms for TOTP as per the [reference implementation](https://tools.ietf.org/html/rfc6238#appendix-A)
-#[derive(Debug, Copy, Clone)]
+#[derive(Debug, Copy, Clone, Eq, PartialEq)]
 #[cfg_attr(feature = "serde_support", derive(Serialize, Deserialize))]
 pub enum Algorithm {
     SHA1,
@@ -117,6 +119,24 @@ pub struct TOTP<T = Vec<u8>> {
     pub secret: T,
 }
 
+impl <T: AsRef<[u8]>> PartialEq for TOTP<T> {
+    fn eq(&self, other: &Self) -> bool {
+        if self.algorithm != other.algorithm {
+            return false;
+        }
+        if self.digits != other.digits {
+            return false;
+        }
+        if self.skew != other.skew {
+            return false;
+        }
+        if self.step != other.step {
+            return false;
+        }
+        constant_time_eq(self.secret.as_ref(), other.secret.as_ref())
+    }
+}
+
 impl<T: AsRef<[u8]>> TOTP<T> {
     /// Will create a new instance of TOTP with given parameters. See [the doc](struct.TOTP.html#fields) for reference as to how to choose those values
     pub fn new(algorithm: Algorithm, digits: usize, skew: u8, step: u64, secret: T) -> TOTP<T> {
@@ -154,7 +174,8 @@ impl<T: AsRef<[u8]>> TOTP<T> {
         let basestep = time / self.step - (self.skew as u64);
         for i in 0..self.skew * 2 + 1 {
             let step_time = (basestep + (i as u64)) * (self.step as u64);
-            if self.generate(step_time) == token {
+            
+            if constant_time_eq(self.generate(step_time).as_bytes(), token.as_bytes()) {
                 return true;
             }
         }
@@ -185,19 +206,78 @@ impl<T: AsRef<[u8]>> TOTP<T> {
     ///
     /// # Errors
     ///
-    /// This will return an error in case the URL gets too long to encode into a QR code
+    /// This will return an error in case the URL gets too long to encode into a QR code.
+    /// This would require the get_url method to generate an url bigger than 2000 characters,
+    /// Which would be too long for some browsers anyway.
     ///
     /// It will also return an error in case it can't encode the qr into a png. This shouldn't happen unless either the qrcode library returns malformed data, or the image library doesn't encode the data correctly
     #[cfg(feature = "qr")]
     pub fn get_qr(&self, label: &str, issuer: &str) -> Result<String, Box<dyn std::error::Error>> {
+        use image::ImageEncoder;
+
         let url = self.get_url(label, issuer);
-        let code = QrCode::new(&url)?;
         let mut vec = Vec::new();
-        let encoder = image::png::PngEncoder::new(&mut vec);
-        encoder.encode(
-            code.render::<Luma<u8>>().build().as_ref(),
-            ((code.width() + 8) * 8) as u32,
-            ((code.width() + 8) * 8) as u32,
+        let qr = qrcodegen::QrCode::encode_text(&url, qrcodegen::QrCodeEcc::Medium)?;
+        let size = qr.size() as u32;
+        
+        // "+ 8 * 8" is here to add padding (the white border around the QRCode)
+        // As some QRCode readers don't work without padding 
+        let image_size = size * 8 + 8 * 8;
+        let mut canvas = image::GrayImage::new(image_size, image_size);
+        
+        // Draw the border
+        for x in 0..image_size {
+            for y in 0..image_size {
+                if y < 8*4 || y >= image_size - 8*4 {
+                    canvas.put_pixel(
+                        x,
+                        y,
+                        Luma([255]),
+                    );
+                    continue;
+                }
+                if x < 8*4 || x >= image_size - 8*4 {
+                    canvas.put_pixel(
+                        x,
+                        y,
+                        Luma([255]),
+                    );
+                }
+            }
+        }
+
+        // The QR inside the white border
+        for x_qr in 0..size {
+            for y_qr in 0..size {
+                // The canvas is a grayscale image without alpha. Hence it's only one 8-bits byte longs
+                // This clever trick to one-line the value was achieved with advanced mathematics
+                // And deep understanding of Boolean algebra.
+                let val = !qr.get_module(x_qr as i32, y_qr as i32) as u8 * 255;
+                
+                // Multiply coordinates by width of pixels
+                // And take into account the 8*4 padding on top and left side
+                let x_start = x_qr * 8 + 8*4;
+                let y_start = y_qr * 8 + 8*4;
+                
+                // Draw a 8-pixels-wide square
+                for x_img in x_start..x_start + 8 {
+                    for y_img in y_start..y_start + 8 {
+                        canvas.put_pixel(
+                            x_img,
+                            y_img,
+                            Luma([val]),
+                        );
+                    }
+                }
+            }
+        }
+
+        // Encode the canvas into a PNG
+        let encoder = image::codecs::png::PngEncoder::new(&mut vec);
+        encoder.write_image(
+            &image::ImageBuffer::from(canvas).into_raw(),
+            image_size,
+            image_size,
             image::ColorType::L8,
         )?;
         Ok(base64::encode(vec))
@@ -208,6 +288,48 @@ impl<T: AsRef<[u8]>> TOTP<T> {
 mod tests {
     use super::*;
 
+    #[test]
+    fn comparison_ok() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        assert_eq!(reference, test);
+    }
+
+    #[test]
+    fn comparison_different_algo() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA256, 6, 1, 1, "TestSecret");
+        assert_ne!(reference, test);
+    }
+
+    #[test]
+    fn comparison_different_digits() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA1, 8, 1, 1, "TestSecret");
+        assert_ne!(reference, test);
+    }
+
+    #[test]
+    fn comparison_different_skew() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA1, 6, 0, 1, "TestSecret");
+        assert_ne!(reference, test);
+    }
+
+    #[test]
+    fn comparison_different_step() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA1, 6, 1, 30, "TestSecret");
+        assert_ne!(reference, test);
+    }
+
+    #[test]
+    fn comparison_different_secret() {
+        let reference = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
+        let test = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecretL");
+        assert_ne!(reference, test);
+    }
+
     #[test]
     fn url_for_secret_matches_sha1() {
         let totp = TOTP::new(Algorithm::SHA1, 6, 1, 1, "TestSecret");
@@ -282,7 +404,7 @@ mod tests {
         let hash_digest = Sha1::digest(qr.as_bytes());
         assert_eq!(
             format!("{:x}", hash_digest).as_str(),
-            "3abc0127e7a2b1013fb25c97ef14422c1fe9e878"
+            "f671a5a553227a9565c6132024808123f2c9e5e3"
         );
     }
 }
